{% extends 'main/layouts/base.html' %}

{% block title %}Welcome to {{ g.__NAME__ }} {{ g.__VERSION__ }} {% endblock %}

{% block content %}
    <div>
        <h1>It's {{ g.__NAME__ }} </h1>
        <h4>v. {{ g.__VERSION__ }}</h4>
    </div>
{% endblock %}
