<!doctype html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <title>{% block title %}{% endblock %}</title>

    {% assets "application_name/styles.css" %}
        <link rel="stylesheet" href="{{ ASSET_URL }}">
    {% endassets %}

  </head>

  <body>
    <header></header>
    <div class="container">
      {% block content %}{% endblock %}
    </div>
    <footer>
      <div>&copy; {{ g.__YEAR__ }} {{ config.APPLICATION_NAME }} {{ config.APPLICATION_VERSION }}</div>
    </footer>

    {% assets "application_name/app.js" %}
      <script type="text/javascript" src="{{ ASSET_URL }}"></script>
    {% endassets %}
  </body>
</html>
